from math import inf
from . import grammar_parser
from .token_trie import TokenTrie, LEAF
from functools import lru_cache
import time
import torch


class LogitsProcessor:
    def __init__(self, grammar):
        self.grammar = grammar
        self.stacks = grammar.init_stacks()
        self.last_size = None

    def accept_token(self, token):
        self.stacks = self.grammar.accept_token(token, self.stacks)

    # TODO: batching
    def __call__(self, input_ids, scores):
        if self.last_size is None:
            pass
        elif len(input_ids[0]) == self.last_size + 1:
            self.stacks = self.grammar.accept_token(input_ids[0][-1], self.stacks)
        else:
            raise "Input size changed"

        # TODO: the <s> token should be accounted for directly rather than just
        # dropped here...
        self.grammar.filter_logits(scores[0], self.stacks)

        self.last_size = len(input_ids[0])
        return scores


class GrammarSampler:
    def __init__(self, input_text, start_rule_name, tokenizer):
        self.tt = 0
        self.nt = 0
        state = grammar_parser.parse(input_text)
        src = state.out_grammar
        self.start_rule_id = state.symbol_ids.get(start_rule_name)

        self.eos_token_id = tokenizer.eos_token_id
        self.token_trie = TokenTrie(tokenizer)
        self.src = src

        pos = 0
        rules = []

        while src[pos] != 0xFFFF:
            rule_id = src[pos]
            if len(rules) <= rule_id:
                rules.extend([None] * (rule_id + 1 - len(rules)))
            rules[rule_id] = pos
            pos += 1
            while src[pos]:
                pos += 1 + src[pos]
            pos += 1

        self.start_rule = rules[self.start_rule_id]
        self.rules = rules

    def logits_processor(self):
        return LogitsProcessor(self)

    def init_stacks(self):
        stack = [self.start_rule + 2]
        return self.advance_stack(tuple(stack))

    @lru_cache(maxsize=8000)
    def advance_stack(self, stack):
        stack = list(stack)
        if len(stack) == 0:
            return [stack]

        pos = stack[-1]

        if self.src[pos] > 1:
            return [stack]

        # The stack head is a nonterminal (a rule reference).
        # Resolving this rule gives a set of one or more possible positions
        # (e.g. two in `a ::= b | c`)
        # We pop the current rule off the stack and, for each option, push:
        # - the symbol following this symbol in the current rule; then
        # - the first symbol of the resolved rule.
        referenced_rule_id = self.src[pos + 1]
        subpos = self.rules[referenced_rule_id] + 1
        stacks = []
        while self.src[subpos]:
            new_stack = stack[:-1]
            if self.src[pos + 2]:
                new_stack.append(pos + 2)
            if self.src[subpos + 1]:
                new_stack.append(subpos + 1)
            stacks.extend(self.advance_stack(tuple(new_stack)))
            subpos += 1 + self.src[subpos]
        return stacks

    def accept(self, byte, stacks):
        new_stacks = []
        for stack in stacks:
            if not stack:
                continue

            pos = stack[-1]
            num_chars = self.src[pos]

            pos += 1
            found = False
            for i in range(0, num_chars, 2):
                if self.src[pos + i] <= byte and byte <= self.src[pos + i + 1]:
                    found = True
                    break
            if not found:
                continue

            pos += num_chars
            new_stack = stack[:-1]
            if self.src[pos]:
                new_stack.append(pos)
            new_stacks.extend(self.advance_stack(tuple(new_stack)))

        return new_stacks

    def accept_token(self, token, stacks):
        if token == self.eos_token_id:
            if any(len(stack) == 0 for stack in stacks):
                return []
            assert False

        for byte in self.token_trie.id2str(token):
            stacks = self.accept(byte, stacks)
            assert stacks != []

        return stacks

    @lru_cache(maxsize=None)
    def pos_char_acceptance(self, pos):
        acceptance = [False] * 256
        num_chars = self.src[pos]
        pos += 1
        for i in range(0, num_chars, 2):
            start = self.src[pos + i]
            end = self.src[pos + i + 1]
            for j in range(start, end + 1):
                acceptance[j] = True
        return acceptance

    @lru_cache(maxsize=1024)
    def token_acceptance_for_stack(self, stack):
        st = time.time()
        stack = list(stack)  # needs to come in as a tuple for lru_cache

        accepts = [False] * len(self.token_trie)
        accepts[self.eos_token_id] = len(stack) == 0

        def traverse_trie(trie, stacks):
            for byte, next_trie in trie.items():
                if byte == LEAF:
                    token_id = next_trie
                    if token_id != self.eos_token_id:
                        accepts[token_id] = bool(stacks)
                    continue

                new_stacks = []
                for stk in stacks:
                    if not stk:
                        continue

                    pos = stk[-1]
                    num_chars = self.src[pos]

                    if not self.pos_char_acceptance(pos)[byte]:
                        continue

                    pos += num_chars + 1
                    new_stack = stk[:-1]
                    if self.src[pos]:
                        new_stack.append(pos)
                    new_stacks.extend(self.advance_stack(tuple(new_stack)))

                if new_stacks:
                    traverse_trie(next_trie, new_stacks)

        traverse_trie(self.token_trie.trie, [stack])

        et = time.time() - st
        x = torch.tensor(accepts, dtype=torch.bool)
        self.tt += et
        self.nt += 1
        return x

    def filter_logits(self, logits, stacks):
        # resolve each stack to a tensor of True/False for each token
        # indicating acceptance
        acceptance = torch.cat(
            [self.token_acceptance_for_stack(tuple(stack)) for stack in stacks]
        )
        # Merge stacks: any True => True
        acceptance = acceptance.reshape(len(stacks), -1).any(dim=0)
        # Logits to -inf where False
        logits[~acceptance] = -inf



import re

LEAF = -1


class TokenTrie:
    def __init__(self, tokenizer):
        self.eos_token_id = tokenizer.eos_token_id
        self.tokens = []
        self.trie = {}
        self.load_tokens(tokenizer)

    def id2str(self, token_id):
        return self.tokens[token_id]

    def __len__(self):
        return len(self.tokens)

    def load_tokens(self, tokenizer):
        def replace_hex(match):
            hex_value = match.group(1)
            return chr(int(hex_value, 16))

        def fmt_token(token):
            token = re.sub(r"<0x([0-9a-fA-F]{2})>", replace_hex, token)
            token = token.replace("▁", " ")
            return bytes(token, "utf-8")

        self.tokens = [
            fmt_token(tokenizer.convert_ids_to_tokens(i))
            for i in range(tokenizer.vocab_size)
        ]
        for token_id, token_bytes in enumerate(self.tokens):
            self.insert_into_trie(self.trie, token_bytes, token_id)

    def insert_into_trie(self, trie, token_bytes, token_id):
        current = trie
        for byte in token_bytes:
            if byte not in current:
                current[byte] = {}
            current = current[byte]
        current[LEAF] = token_id



import sys


class ParseState:
    def __init__(self):
        self.symbol_ids = {}
        self.out_grammar = []


def get_symbol_id(state, src):
    if src not in state.symbol_ids:
        state.symbol_ids[src] = len(state.symbol_ids)
    return state.symbol_ids[src]


def generate_symbol_id(state, base_name):
    next_id = len(state.symbol_ids)
    state.symbol_ids[base_name + "_" + str(next_id)] = next_id
    return next_id


def is_word_char(c):
    return c.isalnum() or c == "-" or c == "_"


def hex_to_int(c):
    if c.isdigit():
        return int(c)
    elif "a" <= c.lower() <= "f":
        return ord(c.lower()) - ord("a") + 10
    return -1


def parse_space(src, newline_ok):
    pos = 0
    while pos < len(src) and (src[pos].isspace() or src[pos] == "#"):
        if src[pos] == "#":
            while pos < len(src) and src[pos] not in ("\r", "\n"):
                pos += 1
        else:
            if not newline_ok and src[pos] in ("\r", "\n"):
                break
            pos += 1
    return src[pos:]


def parse_name(src):
    pos = 0
    while pos < len(src) and is_word_char(src[pos]):
        pos += 1
    if pos == 0:
        raise RuntimeError("expecting name at " + src)
    return src[:pos], src[pos:]


def parse_char(src):
    if src[0] == "\\":
        esc = src[1]
        if esc == "x":
            first = hex_to_int(src[2])
            if first > -1:
                second = hex_to_int(src[3])
                if second > -1:
                    return (first << 4) + second, src[4:]
            raise RuntimeError("expecting \\xNN at " + src)
        elif esc in ('"', "[", "]"):
            return esc, src[2:]
        elif esc == "r":
            return "\r", src[2:]
        elif esc == "n":
            return "\n", src[2:]
        elif esc == "t":
            return "\t", src[2:]
        raise RuntimeError("unknown escape at " + src)
    elif src:
        return src[0], src[1:]
    raise RuntimeError("unexpected end of input")


def parse_sequence(state, src, rule_name, outbuf, is_nested):
    out_start = len(outbuf)

    # sequence size, will be replaced at end when known
    outbuf.append(0)

    last_sym_start = len(outbuf)
    pos = src
    while pos:
        if pos[0] == '"':  # literal string
            pos = pos[1:]
            last_sym_start = len(outbuf)
            while pos[0] != '"':
                char_pair, pos = parse_char(pos)

                # each char of a literal is encoded as a "range" of char - char
                outbuf.append(2)
                outbuf.append(ord(char_pair))
                outbuf.append(ord(char_pair))
            pos = parse_space(pos[1:], is_nested)
        elif pos[0] == "[":  # char range(s)
            pos = pos[1:]
            last_sym_start = len(outbuf)
            # num chars in range - replaced at end of loop
            outbuf.append(0)
            while pos[0] != "]":
                char_pair, pos = parse_char(pos)

                outbuf.append(ord(char_pair))
                if pos[0] == "-" and pos[1] != "]":
                    endchar_pair, pos = parse_char(pos[1:])
                    outbuf.append(ord(endchar_pair))
                else:
                    # chars that aren't part of a c1-c2 range are just doubled (i.e., c-c)
                    outbuf.append(ord(char_pair))
            # replace num chars with actual
            outbuf[last_sym_start] = len(outbuf) - last_sym_start - 1
            pos = parse_space(pos[1:], is_nested)
        elif is_word_char(pos[0]):  # rule reference
            name, pos = parse_name(pos)
            ref_rule_id = get_symbol_id(state, name)
            pos = parse_space(pos, is_nested)
            last_sym_start = len(outbuf)
            outbuf.append(1)
            outbuf.append(ref_rule_id)
        elif pos[0] == "(":  # grouping
            # parse nested alternates into synthesized rule
            pos = parse_space(pos[1:], True)
            sub_rule_id = generate_symbol_id(state, rule_name)
            pos = parse_alternates(state, pos, rule_name, sub_rule_id, True)
            last_sym_start = len(outbuf)
            # output reference to synthesized rule
            outbuf.append(1)
            outbuf.append(sub_rule_id)
            if pos[0] != ")":
                raise RuntimeError("expecting ')' at " + pos)
            pos = parse_space(pos[1:], is_nested)
        elif pos[0] in ("*", "+", "?"):  # repetition operator
            if len(outbuf) - out_start - 1 == 0:
                raise RuntimeError("expecting preceeding item to */+/? at " + pos)
            out_grammar = state.out_grammar

            # apply transformation to previous symbol (last_sym_start -
            # end) according to rewrite rules:
            # S* --> S' ::= S S' |
            # S+ --> S' ::= S S' | S
            # S? --> S' ::= S |
            sub_rule_id = generate_symbol_id(state, rule_name)
            out_grammar.append(sub_rule_id)
            sub_rule_start = len(out_grammar)
            # placeholder for size of 1st alternate
            out_grammar.append(0)
            # add preceding symbol to generated rule
            out_grammar.extend(outbuf[last_sym_start:])
            if pos[0] in ("*", "+"):
                # cause generated rule to recurse
                out_grammar.append(1)
                out_grammar.append(sub_rule_id)
            # apply actual size
            out_grammar[sub_rule_start] = len(out_grammar) - sub_rule_start
            # mark end of 1st alternate
            out_grammar.append(0)
            sub_rule_start = len(out_grammar)
            # placeholder for size of 2nd alternate
            out_grammar.append(0)
            if pos[0] == "+":
                # add preceding symbol as alternate only for '+'
                out_grammar.extend(outbuf[last_sym_start:])
            # apply actual size of 2nd alternate
            out_grammar[sub_rule_start] = len(out_grammar) - sub_rule_start
            # mark end of 2nd alternate, then end of rule
            out_grammar.append(0)
            out_grammar.append(0)

            # in original rule, replace previous symbol with reference to generated rule
            outbuf[last_sym_start:] = [1, sub_rule_id]

            pos = parse_space(pos[1:], is_nested)
        else:
            break
    # apply actual size of this alternate sequence
    outbuf[out_start] = len(outbuf) - out_start
    # mark end of alternate
    outbuf.append(0)
    return pos


def parse_alternates(state, src, rule_name, rule_id, is_nested):
    outbuf = []
    pos = parse_sequence(state, src, rule_name, outbuf, is_nested)
    while pos[0] == "|":
        pos = parse_space(pos[1:], True)
        pos = parse_sequence(state, pos, rule_name, outbuf, is_nested)
    state.out_grammar.append(rule_id)
    state.out_grammar.extend(outbuf)
    state.out_grammar.append(0)
    return pos


def parse_rule(state, src):
    name, pos = parse_name(src)
    pos = parse_space(pos, False)
    rule_id = get_symbol_id(state, name)

    if pos[:3] != "::=":
        raise RuntimeError("expecting ::= at " + pos)
    pos = parse_space(pos[3:], True)

    pos = parse_alternates(state, pos, name, rule_id, False)

    if pos[0] == "\r":
        pos = pos[2:] if pos[1] == "\n" else pos[1:]
    elif pos[0] == "\n":
        pos = pos[1:]
    elif pos:
        raise RuntimeError("expecting newline or end at " + pos)
    return parse_space(pos, True)


def parse(src):
    try:
        state = ParseState()
        pos = parse_space(src, True)
        while pos:
            pos = parse_rule(state, pos)
        state.out_grammar.append(0xFFFF)
        return state
    except RuntimeError as err:
        print("error parsing grammar:", err)
        return ParseState()


def print_rule(file, base, index, symbol_id_names):
    rule_id = base[index]
    print("<{}>{} ::=".format(index, symbol_id_names[rule_id]), end=" ", file=file)
    pos = index + 1
    while base[pos]:
        if pos - 1 > index:
            print("|", end=" ", file=file)
        pos += 1  # sequence size, not needed here
        while base[pos]:
            if base[pos] == 1:
                ref_rule_id = base[pos + 1]
                print(
                    "<{}>{}".format(pos, symbol_id_names[ref_rule_id]),
                    end=" ",
                    file=file,
                )
                pos += 2
            else:
                print("<{}>[".format(pos), end="", file=file)
                num_chars = base[pos]
                pos += 1

                for i in range(0, num_chars, 2):
                    print("{}-".format(chr(base[pos + i])), end="", file=file)
                    if i + 1 < num_chars:
                        print("{}".format(chr(base[pos + i + 1])), end="", file=file)
                print("]", end=" ", file=file)
                pos += num_chars
        pos += 1
    print(file=file)
    return pos + 1


def print_grammar(file, state):
    pos = 0
    symbol_id_names = {v: k for k, v in state.symbol_ids.items()}
    while state.out_grammar[pos] != 0xFFFF:
        pos = print_rule(file, state.out_grammar, pos, symbol_id_names)
    pos = 0
    while state.out_grammar[pos] != 0xFFFF:
        print(f"{state.out_grammar[pos]:04x}", end=" ", file=file)
        pos += 1
    print("ffff\n")


if __name__ == "__main__":
    try:
        with open("examples/grammar.ebnf", "r") as file:
            input_text = file.read()
        state = parse(input_text)
        print_grammar(sys.stdout, state)
        print(state.symbol_ids)
    except FileNotFoundError:
        print("Error: File 'grammar.ebnf' not found.")
    except IOError as e:
        print("Error reading file 'grammar.ebnf':", e)
